
/************* REGULAR CONTEXT SEQUENCE MATCHING  ***********************

File: main.pl
Author: Temur Kutsia
Date: 27.11.2005
The main module of the Implementation of the Context Sequence
Matching Algorithm with regular constraints.

Copyright (c) 2005 Temur Kutsia


=========================================================================

References to the Regular Context sequence Matching Algorithm:

Temur Kutsia and Mircea Marin. Matching with Regular Constraints.
In: Geoff Sutcliffe and Andrei Voronkov, editors. Proc. of 12th
Int. Conference on Logic in Programming, Artificial Intelligence,
and Reasoning (LPAR'05). December 2-6 2005, Montego Bay, Jamaica. 
Volume 3835 of LNAI, pp. 215-229. Springer.

Temur Kutsia and Mircea Marin. Matching with Regular Constraints.
Technical Report 05-05, Research Institute for Symbolic Computation,
Johannes Kepler University of Linz, Austria, 2005.

=========================================================================

Input syntax:
  Context and sequence variables:
      Identifiers that start with lower case letters and end with double
      underscores. Anonymous context and sequence variables are 
      written in single quotes '__'.
      Examples: c__, ctx_or_seq_variable__, vArIaBle1__
  Function and individual variables:
      Identifiers that start with lower case letters and end with a singlr
      underscore.  Anonymous function and individual variables are 
      written in single quotes '_'.
      Examples: i_, fun_or_ind_variable_, vArIaBle1_
  Constants:
       Prolog atoms. Examples: See any textbook on Prolog.
  Reserved words:
       hole, seq. THEY SHOULD NOT BE USED IN THE INPUT PROBLEM.
       star, or, conc. THEY CAN BE USED ONLY AS REGULAR OPERATORS.
  Terms: Built using the grammar:
       T ::= x_ | '_' | x__ | '__' | hole | f(T1,...,Tn) | 
             f_(T1,...,Tn) | c__(T)
       where f is a flexible arity function constant, and T in c__(T)
       is not a sequence variable.
       NOTE: Current implementation does not check input for syntactical
       correctness. It is solely user's responcibility to provide a
       syntactically correct input. In the next versions it will be added.
  Contexts: Terms with only one hole.
  Matching Equations:
       E ::= T1 << T2
       where T1 and T2 are not sequence variables and do not contain hole.
  Regular Expressions on Terms: 
       Built using the grammar:
       R ::= T | seq([]) | (R1,R2) | R1|R2 | R*
       The user can also write conc(R1,R2) for (R1,R2), or(R1,R2) for R1|R2,
       and star(R) for R*.
  Regular Expressions on Contexts: 
       Built using the grammar: 
       Q ::= C | (Q1,Q2) | Q1|Q2 | Q*
       where C is a context.  Also here conc(Q1,Q2) can be written for (Q1,Q2),
       or(Q1,Q2) for Q1|Q2, and star(Q) for Q*.
  Regular Constraints:
       RC ::= [x__, R, 0] | [x__, R, 1] | [c__, Q, 0] | [c__, Q, 1]
       The user can also write [x__, R] for [x__, R, 0] and [c__, Q] for
       [c__, Q, 0].
  Matching problems:
       MP ::= [E1,...,En, RC1,...,RCm].

Output syntax:
  Bindings: 
       B ::= x_ ~> T | x__~> seq([T1,...,Tn]) | f_ ~> f | c__ ~> C
       where n>=0, x_ is a individual variable, x__ is a sequence
       variable, f_ is a function variable, c__ is a context variable,
       Ts are terms, f is a function symbol, C is a context. x=!=T,
       seq([x__])=!=seq([T1,...,Tn]), c__(hole)=!=C.
  Substitutions:
       S ::= [B1,...,Bn]
       n>=0.

************************************************************************/


:-module(main,[csmrc/2,op(120,yfy,~>),op(110,yf,*)]).
:-use_module(rules).
:-use_module(conversions).



                 /**********************************
                 *    csmrc: The main predicate    *
                 ***********************************/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
csmrc(MatchingProblem,Matcher) accepts a matching problem MatchingProblem. If
MatchingProblem is solvable, a matching substitution is returned in Matcher.
The user can get all the answers one by one, by pressing the ';' key after
each answer.

csmrc works as follows: First, the terms in the input problem are converted
into lists of symbols. Then a matcher is computed. Next, if the matcher
contains anonymous variables it is recomputed after instantiating anonymous
variables with fresh singleton variables. From the obtained substitution all
bindings for temporary variables (i.e., those variables that do not occur in
MatchingProblem) are removed. Finally, the output is translated into term
substituions.                  
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

                                       

csmrc(MatchingProblem,Matcher):-
   term_matching_problems_to_symbol_list_matching_problems(MatchingProblem,MP),
   match(MP,[],M),
   recompute_match_with_anonymous_variables_instantiated(MP,M,
                                                         MWithoutAnonymous),
   remove_bindings_for_temporary_variables(MWithoutAnonymous,
                                           SmallerMWithoutAnonymous),
   symbol_list_substitutions_to_term_substitutions(SmallerMWithoutAnonymous,
                                                   Matcher).
