
/************* REGULAR CONTEXT SEQUENCE MATCHING  ***********************

% File: substitutions.pl
% Author: Temur Kutsia
% Date: 27.11.2005
% The module for substitution composition and application.

Copyright (c) 2005 Temur Kutsia

*************************************************************************/

:-module(substitutions,[
          compose/3,
          apply_substitution/3,
          instance/3,
          instance_of_term_list/3,
          op(120,yfy,~>)]).



/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
compose(S1,S2,S) composes two substitutions S1 and S2 and returns S.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

compose(S1,S2,S):-
        apply_substitution(S1,S2,S3),
        remove_duplicated_vars_and_append_substs(S3,S2,S).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
apply_substitution takes as input arguments the lists S1 and S2, and returns
S3. S1 may contain bindings, matching equations, regular constraints, and
terms. S2 is a substitution. S3 is obtained from S1 by applying S2 to
terms in it (except variables X in regular constraints [X,R,F]). 
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

apply_substitution([],_,[]):-!.
apply_substitution(S,[],S):-!.
apply_substitution([X ~> Y |Tail],S,[X ~> Y1| Tail1]):-!,
        instance(Y,S,Y1),
        apply_substitution(Tail, S, Tail1).
apply_substitution([X << Y |Tail],S,[X1 << Y| Tail1]):- !,
        instance(X,S,X1),
        apply_substitution(Tail, S, Tail1).
apply_substitution([[X,Regexp,Flag]|Tail],S,[[X,Regexp1,Flag1]|Tail1]):-!,
        instance(Regexp,S,Regexp1),
        instance(Flag,S,Flag1),
        apply_substitution(Tail, S, Tail1).
apply_substitution([X |Tail],S,[X1 | Tail1]):-
        instance(X,S,X1),
        apply_substitution(Tail, S, Tail1).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
instance(X,Subst,I) returns an instance I of X with respect to the 
substitution Subst.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

instance(circplus(X,Y),Subst,circplus(XInst,YInst)):-!,
        instance(X,Subst,XInst),
        instance(Y,Subst,YInst).
instance(TermSequence_or_Regexp_or_Term,Subst,Instance):-
        instance_of_sequence(TermSequence_or_Regexp_or_Term,Subst,Instance),
        !.
instance(TermSequence_or_Regexp_or_Term,Subst,Instance):-
        instance_of_regexp(TermSequence_or_Regexp_or_Term,Subst,Instance),
        !.
instance(TermSequence_or_Regexp_or_Term,Subst,Instance):-
        instance_of_nonseqvar_term(TermSequence_or_Regexp_or_Term,Subst,
                                   Instance).


% Instance of seq expressions.

instance_of_sequence(seq(L),Subst,seq(Instance)):-
        instance_of_term_list(L,Subst,LInst),
        flatten_sequence(LInst,Instance).


% Instance of regular expressions.

instance_of_regexp(star(R),Subst,star(Instance)):-!,
        instance_of_regexp(R,Subst,Instance).
instance_of_regexp(or(R1,R2),Subst,or(R1Inst,R2Inst)):-!,
        instance_of_regexp(R1,Subst,R1Inst),
        instance_of_regexp(R2,Subst,R2Inst).
instance_of_regexp(conc(R1,R2),Subst,conc(R1Inst,R2Inst)):-!,
        instance_of_regexp(R1,Subst,R1Inst),
        instance_of_regexp(R2,Subst,R2Inst).
instance_of_regexp(seq([]),_,seq([])):-!.
instance_of_regexp(hole,_,hole):-!.
instance_of_regexp([ctx_or_seq_var(CtxVar),hole],Subst,Context):-!,
        instance_of_ctx_var(ctx_or_seq_var(CtxVar),Subst,Context).
instance_of_regexp(ctx_or_seq_var(SeqVar),Subst,Conc):-
 %       SeqVar\=[_|_],        % what is this?
        !,
        instance_of_symbol(ctx_or_seq_var(SeqVar),Subst,Sequence),
        seq_to_conc(Sequence,Conc).
instance_of_regexp(Term,Subst,Instance):-
        instance_of_nonseqvar_term(Term,Subst,Instance).


% Instance of terms that are not sequence variables

instance_of_nonseqvar_term([ctx_or_seq_var(CtxVar),Arg],Subst,Instance):-
        !,
        instance_of_ctx_var(ctx_or_seq_var(CtxVar),Subst,Context),
        instance_of_nonseqvar_term(Arg,Subst,ArgInst),
        apply_context(Context,ArgInst,Instance).

instance_of_nonseqvar_term([FunVar_or_FunSymb|ArgList],Subst,
                           [FunSymb|ArgInst]):-
        !,
        instance_of_symbol(FunVar_or_FunSymb,Subst,FunSymb),
        instance_of_term_list(ArgList,Subst,ArgInst).

instance_of_nonseqvar_term(Symbol,Subst,Instance):-
        instance_of_symbol(Symbol,Subst,Instance).


% Instance of a list of terms.

instance_of_term_list([ctx_or_seq_var(SeqVar)|OtherTerms],Subst,Instance):-
 %       SeqVar \= [_|_],   % do we still need it?
        !,
        instance_of_symbol(ctx_or_seq_var(SeqVar),Subst,Sequence),
        instance_of_term_list(OtherTerms,Subst,OtherTermsInst),
        flatten_sequence([Sequence|OtherTermsInst],Instance).

instance_of_term_list([Term|OtherTerms],Subst,
                      [TermInstance|OtherTermsInst]):-
        instance_of_nonseqvar_term(Term,Subst,TermInstance),
        instance_of_term_list(OtherTerms,Subst,OtherTermsInst).

instance_of_term_list([],_,[]).


% Instance of a context variable.

instance_of_ctx_var(ctx_or_seq_var(CtxVar),Subst,Instance):-
        member(ctx_or_seq_var(CtxVar)~>Instance,Subst),
        !.
instance_of_ctx_var(ctx_or_seq_var(CtxVar),_,[ctx_or_seq_var(CtxVar),hole]).


% Instance of a symbol.

instance_of_symbol(Var,Subst,Instance):-
        member(Var~>Instance,Subst),
        !.
instance_of_symbol(Constant_or_var_not_in_dom,_,
                   Constant_or_var_not_in_dom).


% Transforming seq expressions into concatenation.

seq_to_conc(seq([X]),X):-!.
seq_to_conc(seq([X,Y|T]),conc(X,R)):-seq_to_conc(seq([Y|T]),R).
seq_to_conc(seq([]),seq([])).


% Context application.

apply_context(hole,Term,Term):-!.
apply_context([hole],Term,Term):-!.
apply_context([Term1,hole],Term,[Term1,Term]):-!.
apply_context([hole|CtxTail],Term,[Term|CtxTail]):-!.
apply_context([[CxtHead|CtxTail]|Tail],Term,[NewTerm|Tail]):-
        apply_context([CxtHead|CtxTail],Term,NewTerm).
apply_context([Head|Tail],Term,[Head|NewTail]):-
        apply_context(Tail,Term,NewTail).

% Flattening nested seq expressions.

flatten_sequence([seq([])|Tail],Tail):-!.
flatten_sequence([seq([X|T])|Tail],[X|Flat]):-!,
        flatten_sequence([seq(T)|Tail],Flat).
flatten_sequence([X|Tail],[X1|Flat]):-!,
        flatten_sequence(X,X1),
        flatten_sequence(Tail,Flat).
flatten_sequence(X,X).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
remove_duplicated_vars_and_append_substs(S1,S2,S) takes two lists of bindings
S1 and S2, deletes all the bindings X~>T from S2 such that S1 contains a
binding for X, deletes all the bindings of the form X~>X from S1, and joins
the obtained lists together into S. 
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


remove_duplicated_vars_and_append_substs([],L,L):-!.
remove_duplicated_vars_and_append_substs(L,[],L):-!.
remove_duplicated_vars_and_append_substs(Subst1,Subst2,Result):-
        remove_vars_from_second_that_occur_in_first(Subst1,Subst2,
                                                    [],Subst2New),
        remove_identities(Subst1,Subst1New),
        append(Subst1New,Subst2New,Result).
remove_vars_from_second_that_occur_in_first([X~>_|Rest1],[X~>_|Rest2],
                                            ResultSoFar,Result):-
        !,
        remove_vars_from_second_that_occur_in_first(Rest1,Rest2,
                                                    ResultSoFar,Result).
remove_vars_from_second_that_occur_in_first([X~>Y|Rest1],[B|Rest2],
                                            Cleaned,Result):-
        !,
        remove_vars_from_second_that_occur_in_first([X~>Y|Rest1],Rest2,
                                                    [B|Cleaned],Result).
remove_vars_from_second_that_occur_in_first([_|Rest1],[],CleanedOfFirstVar,
                                            Result):-!,
        remove_vars_from_second_that_occur_in_first(Rest1,CleanedOfFirstVar,
                                                    [],Result).
remove_vars_from_second_that_occur_in_first([],Result,_,Result).

remove_identities([],[]).
remove_identities([X~>X|Rest],Result):-
        !,
        remove_identities(Rest,Result).
remove_identities([B|Rest],[B|Result]):-
        remove_identities(Rest,Result).





